BASE_URL = "http://api.nbp.pl/api" 

LATEST_CURRENCY_RATE_URL = lambda table, currency: f"{BASE_URL}/exchangerates/rates/{table}/{currency}/"
TODAY_CURRENCY_RATE_URL = lambda table, currency: f"{BASE_URL}/exchangerates/rates/{table}/{currency}/today"
CURRENCY_RATE_FROM_DATE_URL = lambda table, currency, date: f"{BASE_URL}/exchangerates/rates/{table}/{currency}/{date}/"
CURRENCY_RATE_FROM_START_DATE_TO_END_DATE_URL = lambda table, currency, start_date, end_date: f"{BASE_URL}/exchangerates/rates/{table}/{currency}/{start_date}/{end_date}/"

# gold
LATEST_GOLD_RATE_URL = f"{BASE_URL}/cenyzlota"
TODAY_GOLD_RATE_URL = f"{BASE_URL}/cenyzlota/today"
GOLD_RATE_FROM_DATE_URL = lambda date: f"{BASE_URL}/cenyzlota/{date}"
GOLD_RATE_FROM_START_DATE_TO_END_DATE_URL = lambda start_date, end_date: f"{BASE_URL}/cenyzlota/{start_date}/{end_date}"



from enum import Enum


class RequestType(Enum):
    GET = "GET"


class CurrencyType(Enum):
    USD = "USD"
    AUD = "AUD"
    HKD = "HKD"
    CAD = "CAD"
    SGD = "SGD"
    EUR = "EUR"
    CHF = "CHF"
    GBP = "GBP"
    JPY = "JPY"
    CZK = "CZK"
    NOK = "NOK"
    SEK = "SEK"
    KRW = "KRW"
    HUF = "HUF"
    BRL = "BRL"


class TableType(Enum):
    A = "A"
    B = "B"
    C = "C"



from datetime import datetime, timedelta

from ._request import ApiRequester
from .types import CurrencyType, RequestType
from ._utils import get_table
from ._exceptions import NotExistsError
from .models import Currency, Gold

from . import _urls
from . import _settings


class _NBPApiClientBase:
    def raw_request(self, url: str, type: RequestType, params: dict | None = None) -> dict | None:
        """
        Umozliwia bezposrednie odpytanie API
        """
        match type:
            case RequestType.GET:
                return ApiRequester(url=url).get(params=params)


class _NBPApiClientGoldPriceMixin(_NBPApiClientBase):
    def get_gold_rate_currently_valid(self) -> Gold | None:
        """
        Aktualnie obowiązująca cena złota
        """

        try:
            response = ApiRequester(
                url=_urls.LATEST_GOLD_RATE_URL
            ).get()
        except NotExistsError:
            return

        return Gold.from_response(response[0])

    def get_gold_rate_from_today(self) -> Gold | None:
        """
        Cena złota opublikowana w dniu dzisiejszym (albo brak danych)
        """
        try:
            response = ApiRequester(
                url=_urls.TODAY_GOLD_RATE_URL
            ).get()
        except NotExistsError:
            return
        
        return Gold.from_response(response[0])
    
    def get_gold_rate_from_date(self, date: datetime) -> Gold | None:
        """
        Cena złota opublikowana w dniu {date} (albo brak danych)
        """
        try:
            response = ApiRequester(
                url=_urls.GOLD_RATE_FROM_DATE_URL(date.strftime(_settings.DATE_FORMAT))
            ).get()
        except NotExistsError:
            return
        
        return Gold.from_response(response[0])

    def get_gold_rate_from_working_day_before_date(self, date: datetime) -> Gold:
        """
        Cena złota opublikowana w dniu roboczym poprzedzającym podaną datę {date}

        Jeśli podasz datę której dzień tygodnia wypada w poniedziałek, kurs zostanie pobrany z piątku.
        Jeśli podasz datę której dzień tygodnia wypada w środę, kurs zostanie pobrany z wtorku.
        """
        response = None
        before_date = date - timedelta(days=1)

        while True:
            try:
                response = ApiRequester(
                    url=_urls.GOLD_RATE_FROM_DATE_URL(
                        before_date.strftime(_settings.DATE_FORMAT)
                    )
                ).get()
            except NotExistsError:
                before_date = before_date - timedelta(days=1)
                continue
            else:
                break

        return Gold.from_response(response[0])

    def get_gold_rate_from_start_date_to_end_date(
            self,
            start_date: datetime,
            end_date: datetime
        ) -> list[Gold]:
        """
        Seria tabel kursów typu opublikowanych w zakresie dat od {start_date} do {end_date} (albo brak danych)
        """
        try:
            response = ApiRequester(
                url=_urls.GOLD_RATE_FROM_START_DATE_TO_END_DATE_URL(
                    start_date.strftime(_settings.DATE_FORMAT),
                    end_date.strftime(_settings.DATE_FORMAT)
                )
            ).get()
        except NotExistsError:
            return []
        
        return [Gold.from_response(gold) for gold in response]


class _NBPApiClientCurrencyMixin(_NBPApiClientBase):
    def get_currency_rate_currently_valid(
            self,
            currency: CurrencyType
        ) -> Currency | None:
        """
        Aktualnie obowiązujący kurs waluty w PLN
        """
        table = get_table(currency)

        try:
            response = ApiRequester(
                url=_urls.LATEST_CURRENCY_RATE_URL(table.value, currency.value)
            ).get()
        except NotExistsError:
            return

        return Currency.from_response(response["rates"][0], type=currency)

    def get_currency_rate_from_today(
            self,
            currency: CurrencyType
        ) -> Currency | None:
        """
        Kurs waluty w PLN opublikowany w dniu dzisiejszym (albo brak danych)
        """
        table = get_table(currency)

        try:
            response = ApiRequester(
                url=_urls.TODAY_CURRENCY_RATE_URL(table.value, currency.value)
            ).get()
        except NotExistsError:
            return

        return Currency.from_response(response["rates"][0], type=currency)
    
    def get_currency_rate_from_date(
            self,
            currency: CurrencyType,
            date: datetime
        ) -> Currency | None:
        """
        Kurs waluty opublikowany w dniu {date} (albo brak danych)
        """
        table = get_table(currency)

        try:
            response = ApiRequester(
                url=_urls.CURRENCY_RATE_FROM_DATE_URL(
                    table.value,
                    currency.value,
                    date.strftime(_settings.DATE_FORMAT)
                )
            ).get()
        except NotExistsError:
            return
    
        return Currency.from_response(response["rates"][0], type=currency)

    def get_currency_rate_from_working_day_before_date(
            self,
            currency: CurrencyType,
            date: datetime
        ) -> Currency:
        """
        Kurs waluty opublikowany w dniu roboczym poprzedzającym podaną datę {date}

        Jeśli podasz datę której dzień tygodnia wypada w poniedziałek, kurs zostanie pobrany z piątku.
        Jeśli podasz datę której dzień tygodnia wypada w środę, kurs zostanie pobrany z wtorku.
        """
        table = get_table(currency)

        response = None
        before_date = date - timedelta(days=1)

        while True:
            try:
                response = ApiRequester(
                    url=_urls.CURRENCY_RATE_FROM_DATE_URL(
                        table.value,
                        currency.value,
                        before_date.strftime(_settings.DATE_FORMAT)
                    )
                ).get()
            except NotExistsError:
                before_date = before_date - timedelta(days=1)
                continue
            else:
                break

        return Currency.from_response(response["rates"][0], type=currency)

    def get_currency_rate_from_start_date_to_end_date(
            self,
            currency: CurrencyType,
            start_date: datetime,
            end_date: datetime
        ) -> list[Currency]:
        """
        Seria tabel kursów typu opublikowanych w zakresie dat od {start_date} do {end_date} (albo brak danych)
        """
        table = get_table(currency)

        try:
            response = ApiRequester(
                url=_urls.CURRENCY_RATE_FROM_START_DATE_TO_END_DATE_URL(
                    table.value,
                    currency.value,
                    start_date.strftime(_settings.DATE_FORMAT),
                    end_date.strftime(_settings.DATE_FORMAT)
                )
            ).get()
        except NotExistsError:
            return []

        return [Currency.from_response(rate, type=currency) for rate in response["rates"]]


class NBPApiClient(
    _NBPApiClientCurrencyMixin,
    _NBPApiClientGoldPriceMixin
):
    pass


import requests

from ._exceptions import ApiError, NotExistsError


class ApiRequester:
    def __init__(self, url: str):
        self.url = url

    def get(self, params: dict | None = None) -> dict:
        if params:
            params = {k: v for k, v in params.items() if v} 
        
        response = requests.get(self.url, params)

        if response.status_code == 404:
            raise NotExistsError
        if response.status_code >= 300:
            raise ApiError(response.json()["error"])
        
        return response.json()
 


from .types import CurrencyType, TableType


def get_table(currency: CurrencyType) -> TableType | None:
    mapper = {
        TableType.A: (
            CurrencyType.USD,
            CurrencyType.AUD,
            CurrencyType.HKD,
            CurrencyType.CAD,
            CurrencyType.SGD,
            CurrencyType.EUR,
            CurrencyType.CHF,
            CurrencyType.GBP,
            CurrencyType.JPY,
            CurrencyType.CZK,
            CurrencyType.NOK,
            CurrencyType.SEK,
            CurrencyType.KRW,
            CurrencyType.HUF,
            CurrencyType.BRL
        )
    }

    for table, currencies in mapper.items():
        if currency in currencies:
            return table



from unittest import TestCase
from datetime import datetime, timedelta

from src.nbp_sdk.client import NBPApiClient
from src.nbp_sdk.types import CurrencyType


class TestGoldNBPApiClient(TestCase):
    def setUp(self):
        self.api = NBPApiClient()

    # def test_gold_simple(self):
    #     assert self.api.get_gold_rate_currently_valid()
    #     assert self.api.get_gold_rate_from_today()


    def test_gold_rate_from_date(self):
        date = datetime(2024, 3, 14)
 
        gold = self.api.get_gold_rate_from_date(date)
        assert gold
        assert gold.effective_date == date

    def test_gold_rate_from_working_day_before_date(self):
        date_monday = datetime(2024, 3, 18)

        gold = self.api.get_gold_rate_from_working_day_before_date(date_monday)
        assert gold
        assert gold.effective_date == date_monday - timedelta(days=3)

    def test_gold_rate_from_start_date_to_end_date__weekend(self):
        start_date = datetime(2024, 3, 14)
        end_date = datetime(2024, 3, 17) 
        golds = self.api.get_gold_rate_from_start_date_to_end_date(start_date, end_date)
        assert len(golds) == 2

        for index, gold in enumerate(golds):
            assert gold.effective_date == start_date + timedelta(days=index)

    def test_gold_rate_from_start_date_to_end_date(self):
        start_date = datetime(2024, 3, 11)
        end_date = datetime(2024, 3, 14) 
        golds = self.api.get_gold_rate_from_start_date_to_end_date(start_date, end_date)
        assert len(golds) == 4

        for index, gold in enumerate(golds):
            assert gold.effective_date == start_date + timedelta(days=index)


class TestCurrenciesNBPApiClient(TestCase):
    def setUp(self):
        self.api = NBPApiClient()

    # def test_currency_simple(self):
    #     assert self.api.get_currency_rate_currently_valid(CurrencyType.USD)
    #     assert self.api.get_currency_rate_from_today(CurrencyType.EUR)

    def test_currenct_rate_from_date(self):
        date = datetime(2024, 3, 14)
 
        currency = self.api.get_currency_rate_from_date(CurrencyType.USD, date)
        assert currency
        assert currency.effective_date == date

    def test_currency_rate_from_working_day_before_date(self):
        date_monday = datetime(2024, 3, 18)

        currency = self.api.get_currency_rate_from_working_day_before_date(CurrencyType.USD, date_monday)
        assert currency
        assert currency.effective_date == date_monday - timedelta(days=3)

    def test_currency_rate_from_start_date_to_end_date__weekend(self):
        start_date = datetime(2024, 3, 14)
        end_date = datetime(2024, 3, 17) 
        currencies = self.api.get_currency_rate_from_start_date_to_end_date(CurrencyType.USD, start_date, end_date)
        assert len(currencies) == 2

        for index, currency in enumerate(currencies):
            assert currency.effective_date == start_date + timedelta(days=index)

    def test_currency_rate_from_start_date_to_end_date(self):
        start_date = datetime(2024, 3, 11)
        end_date = datetime(2024, 3, 14) 
        currencies = self.api.get_currency_rate_from_start_date_to_end_date(CurrencyType.AUD, start_date, end_date)
        assert len(currencies) == 4

        for index, currency in enumerate(currencies):
            assert currency.effective_date == start_date + timedelta(days=index)



from dataclasses import dataclass
from decimal import Decimal
from datetime import datetime

from .types import CurrencyType
from . import _settings


@dataclass
class Currency:
    type: CurrencyType
    # przeliczony kurs średni waluty
    average_rate: Decimal
    table_no: str
    # data publikacji
    effective_date: datetime

    @classmethod
    def from_response(cls, d: dict, type: CurrencyType):
        return cls(
            average_rate=Decimal(str(d["mid"])),
            effective_date=datetime.strptime(d["effectiveDate"], _settings.DATE_FORMAT),
            table_no=d["no"],
            type=type,
        )


@dataclass
class Gold:
    # wyliczona w NBP cena 1 g złota (w próbie 1000)
    rate_per_1g: Decimal
    # data publikacji
    effective_date: datetime

    @classmethod
    def from_response(cls, d: dict):
        return cls(
           rate_per_1g=Decimal(str(d["cena"])),
           effective_date=datetime.strptime(d["data"], _settings.DATE_FORMAT), 
        )
